@@ -21,19 +21,26 @@ from functools import cached_property
 import typing as T
 import os, sys
 
-import gi # type: ignore
+import gi  # type: ignore
+
 gi.require_version("GIRepository", "2.0")
-from gi.repository import GIRepository # type: ignore
+from gi.repository import GIRepository  # type: ignore
 
 from .errors import CompileError, CompilerBugError
 from . import typelib, xml_reader
 
-_namespace_cache = {}
+_namespace_cache: T.Dict[str, "Namespace"] = {}
 _xml_cache = {}
 
+_user_search_paths = []
 
-def get_namespace(namespace, version):
-    search_paths = GIRepository.Repository.get_search_path()
+
+def add_typelib_search_path(path: str):
+    _user_search_paths.append(path)
+
+
+def get_namespace(namespace: str, version: str) -> "Namespace":
+    search_paths = [*GIRepository.Repository.get_search_path(), *_user_search_paths]
 
     filename = f"{namespace}-{version}.typelib"
 
@@ -57,13 +64,13 @@ def get_namespace(namespace, version):
     return _namespace_cache[filename]
 
 
-def get_xml(namespace, version):
-    from .main import VERSION
-    from xml.etree import ElementTree
+def get_xml(namespace: str, version: str):
     search_paths = []
 
     if data_paths := os.environ.get("XDG_DATA_DIRS"):
-        search_paths += [os.path.join(path, "gir-1.0") for path in data_paths.split(os.pathsep)]
+        search_paths += [
+            os.path.join(path, "gir-1.0") for path in data_paths.split(os.pathsep)
+        ]
 
     filename = f"{namespace}-{version}.gir"
 
@@ -86,16 +93,68 @@ def get_xml(namespace, version):
 
 class GirType:
     @property
-    def doc(self):
+    def doc(self) -> T.Optional[str]:
         return None
 
-    def assignable_to(self, other) -> bool:
+    def assignable_to(self, other: "GirType") -> bool:
+        raise NotImplementedError()
+
+    @property
+    def name(self) -> str:
+        """The GIR name of the type, not including the namespace"""
         raise NotImplementedError()
 
     @property
     def full_name(self) -> str:
+        """The GIR name of the type to use in diagnostics"""
         raise NotImplementedError()
 
+    @property
+    def glib_type_name(self) -> str:
+        """The name of the type in the GObject type system, suitable to pass to `g_type_from_name()`."""
+        raise NotImplementedError()
+
+    @property
+    def incomplete(self) -> bool:
+        return False
+
+
+class ExternType(GirType):
+    def __init__(self, name: str) -> None:
+        super().__init__()
+        self._name = name
+
+    def assignable_to(self, other: GirType) -> bool:
+        return True
+
+    @property
+    def full_name(self) -> str:
+        return self._name
+
+    @property
+    def glib_type_name(self) -> str:
+        return self._name
+
+    @property
+    def incomplete(self) -> bool:
+        return True
+
+
+class ArrayType(GirType):
+    def __init__(self, inner: GirType) -> None:
+        self._inner = inner
+
+    def assignable_to(self, other: GirType) -> bool:
+        return isinstance(other, ArrayType) and self._inner.assignable_to(other._inner)
+
+    @property
+    def name(self) -> str:
+        return self._inner.name + "[]"
+
+    @property
+    def full_name(self) -> str:
+        return self._inner.full_name + "[]"
+
 
 class BasicType(GirType):
     name: str = "unknown type"
@@ -104,53 +163,85 @@ class BasicType(GirType):
     def full_name(self) -> str:
         return self.name
 
+
 class BoolType(BasicType):
     name = "bool"
-    def assignable_to(self, other) -> bool:
+    glib_type_name: str = "gboolean"
+
+    def assignable_to(self, other: GirType) -> bool:
         return isinstance(other, BoolType)
 
+
 class IntType(BasicType):
     name = "int"
-    def assignable_to(self, other) -> bool:
-        return isinstance(other, IntType) or isinstance(other, UIntType) or isinstance(other, FloatType)
+    glib_type_name: str = "gint"
+
+    def assignable_to(self, other: GirType) -> bool:
+        return (
+            isinstance(other, IntType)
+            or isinstance(other, UIntType)
+            or isinstance(other, FloatType)
+        )
+
 
 class UIntType(BasicType):
     name = "uint"
-    def assignable_to(self, other) -> bool:
-        return isinstance(other, IntType) or isinstance(other, UIntType) or isinstance(other, FloatType)
+    glib_type_name: str = "guint"
+
+    def assignable_to(self, other: GirType) -> bool:
+        return (
+            isinstance(other, IntType)
+            or isinstance(other, UIntType)
+            or isinstance(other, FloatType)
+        )
+
 
 class FloatType(BasicType):
     name = "float"
-    def assignable_to(self, other) -> bool:
+    glib_type_name: str = "gfloat"
+
+    def assignable_to(self, other: GirType) -> bool:
         return isinstance(other, FloatType)
 
+
 class StringType(BasicType):
     name = "string"
-    def assignable_to(self, other) -> bool:
+    glib_type_name: str = "gchararray"
+
+    def assignable_to(self, other: GirType) -> bool:
         return isinstance(other, StringType)
 
+
+class TypeType(BasicType):
+    name = "GType"
+    glib_type_name: str = "GType"
+
+    def assignable_to(self, other: GirType) -> bool:
+        return isinstance(other, TypeType)
+
+
 _BASIC_TYPES = {
-    "gboolean": BoolType,
+    "bool": BoolType,
+    "string": StringType,
     "int": IntType,
-    "gint": IntType,
-    "gint64": IntType,
-    "guint": UIntType,
-    "guint64": UIntType,
-    "gfloat": FloatType,
-    "gdouble": FloatType,
+    "uint": UIntType,
     "float": FloatType,
     "double": FloatType,
-    "utf8": StringType,
+    "type": TypeType,
 }
 
+
+TNode = T.TypeVar("TNode", bound="GirNode")
+
+
 class GirNode:
-    def __init__(self, container, tl):
+    def __init__(self, container: T.Optional["GirNode"], tl: typelib.Typelib) -> None:
         self.container = container
         self.tl = tl
 
-    def get_containing(self, container_type):
+    def get_containing(self, container_type: T.Type[TNode]) -> TNode:
         if self.container is None:
-            return None
+            raise CompilerBugError()
         elif isinstance(self.container, container_type):
             return self.container
         else:
@@ -163,11 +254,11 @@ class GirNode:
                 return el
 
     @cached_property
-    def glib_type_name(self):
+    def glib_type_name(self) -> str:
         return self.tl.OBJ_GTYPE_NAME
 
     @cached_property
-    def full_name(self):
+    def full_name(self) -> str:
         if self.container is None:
             return self.name
         else:
@@ -208,20 +299,16 @@ class GirNode:
         return None
 
     @property
-    def type_name(self):
-        return self.type.name
-
-    @property
-    def type(self):
-        return self.get_containing(Namespace).lookup_type(self.type_name)
+    def type(self) -> GirType:
+        raise NotImplementedError()
 
 
 class Property(GirNode):
-    def __init__(self, klass, tl: typelib.Typelib):
+    def __init__(self, klass: T.Union["Class", "Interface"], tl: typelib.Typelib):
         super().__init__(klass, tl)
 
     @cached_property
-    def name(self):
+    def name(self) -> str:
         return self.tl.PROP_NAME
 
     @cached_property
@@ -230,24 +317,26 @@ class Property(GirNode):
 
     @cached_property
     def signature(self):
-        return f"{self.type_name} {self.container.name}.{self.name}"
+        return f"{self.full_name} {self.container.name}.{self.name}"
 
     @property
-    def writable(self):
+    def writable(self) -> bool:
         return self.tl.PROP_WRITABLE == 1
 
     @property
-    def construct_only(self):
+    def construct_only(self) -> bool:
         return self.tl.PROP_CONSTRUCT_ONLY == 1
 
 
 class Parameter(GirNode):
-    def __init__(self, container: GirNode, tl: typelib.Typelib):
+    def __init__(self, container: GirNode, tl: typelib.Typelib) -> None:
         super().__init__(container, tl)
 
 
 class Signal(GirNode):
-    def __init__(self, klass, tl: typelib.Typelib):
+    def __init__(
+        self, klass: T.Union["Class", "Interface"], tl: typelib.Typelib
+    ) -> None:
         super().__init__(klass, tl)
         # if parameters := xml.get_elements('parameters'):
         #     self.params = [Parameter(self, child) for child in parameters[0].get_elements('parameter')]
@@ -263,11 +352,11 @@ class Signal(GirNode):
 
 
 class Interface(GirNode, GirType):
-    def __init__(self, ns, tl: typelib.Typelib):
+    def __init__(self, ns: "Namespace", tl: typelib.Typelib):
         super().__init__(ns, tl)
 
     @cached_property
-    def properties(self):
+    def properties(self) -> T.Mapping[str, Property]:
         n_prerequisites = self.tl.INTERFACE_N_PREREQUISITES
         offset = self.tl.header.HEADER_INTERFACE_BLOB_SIZE
         offset += (n_prerequisites + n_prerequisites % 2) * 2
@@ -280,11 +369,13 @@ class Interface(GirNode, GirType):
         return result
 
     @cached_property
-    def signals(self):
+    def signals(self) -> T.Mapping[str, Signal]:
         n_prerequisites = self.tl.INTERFACE_N_PREREQUISITES
         offset = self.tl.header.HEADER_INTERFACE_BLOB_SIZE
         offset += (n_prerequisites + n_prerequisites % 2) * 2
-        offset += self.tl.INTERFACE_N_PROPERTIES * self.tl.header.HEADER_PROPERTY_BLOB_SIZE
+        offset += (
+            self.tl.INTERFACE_N_PROPERTIES * self.tl.header.HEADER_PROPERTY_BLOB_SIZE
+        )
         offset += self.tl.INTERFACE_N_METHODS * self.tl.header.HEADER_FUNCTION_BLOB_SIZE
         n_signals = self.tl.INTERFACE_N_SIGNALS
         property_size = self.tl.header.HEADER_SIGNAL_BLOB_SIZE
@@ -295,7 +386,7 @@ class Interface(GirNode, GirType):
         return result
 
     @cached_property
-    def prerequisites(self):
+    def prerequisites(self) -> T.List["Interface"]:
         n_prerequisites = self.tl.INTERFACE_N_PREREQUISITES
         result = []
         for i in range(n_prerequisites):
@@ -303,7 +394,7 @@ class Interface(GirNode, GirType):
             result.append(self.get_containing(Repository)._resolve_dir_entry(entry))
         return result
 
-    def assignable_to(self, other) -> bool:
+    def assignable_to(self, other: GirType) -> bool:
         if self == other:
             return True
         for pre in self.prerequisites:
@@ -313,15 +404,15 @@ class Interface(GirNode, GirType):
 
 
 class Class(GirNode, GirType):
-    def __init__(self, ns, tl: typelib.Typelib):
+    def __init__(self, ns: "Namespace", tl: typelib.Typelib) -> None:
         super().__init__(ns, tl)
 
     @property
-    def abstract(self):
+    def abstract(self) -> bool:
         return self.tl.OBJ_ABSTRACT == 1
 
     @cached_property
-    def implements(self):
+    def implements(self) -> T.List[Interface]:
         n_interfaces = self.tl.OBJ_N_INTERFACES
         result = []
         for i in range(n_interfaces):
@@ -330,12 +421,14 @@ class Class(GirNode, GirType):
         return result
 
     @cached_property
-    def own_properties(self):
+    def own_properties(self) -> T.Mapping[str, Property]:
         n_interfaces = self.tl.OBJ_N_INTERFACES
         offset = self.tl.header.HEADER_OBJECT_BLOB_SIZE
         offset += (n_interfaces + n_interfaces % 2) * 2
         offset += self.tl.OBJ_N_FIELDS * self.tl.header.HEADER_FIELD_BLOB_SIZE
-        offset += self.tl.OBJ_N_FIELD_CALLBACKS * self.tl.header.HEADER_CALLBACK_BLOB_SIZE
+        offset += (
+            self.tl.OBJ_N_FIELD_CALLBACKS * self.tl.header.HEADER_CALLBACK_BLOB_SIZE
+        )
         n_properties = self.tl.OBJ_N_PROPERTIES
         property_size = self.tl.header.HEADER_PROPERTY_BLOB_SIZE
         result = {}
@@ -345,12 +438,14 @@ class Class(GirNode, GirType):
         return result
 
     @cached_property
-    def own_signals(self):
+    def own_signals(self) -> T.Mapping[str, Signal]:
         n_interfaces = self.tl.OBJ_N_INTERFACES
         offset = self.tl.header.HEADER_OBJECT_BLOB_SIZE
         offset += (n_interfaces + n_interfaces % 2) * 2
         offset += self.tl.OBJ_N_FIELDS * self.tl.header.HEADER_FIELD_BLOB_SIZE
-        offset += self.tl.OBJ_N_FIELD_CALLBACKS * self.tl.header.HEADER_CALLBACK_BLOB_SIZE
+        offset += (
+            self.tl.OBJ_N_FIELD_CALLBACKS * self.tl.header.HEADER_CALLBACK_BLOB_SIZE
+        )
         offset += self.tl.OBJ_N_PROPERTIES * self.tl.header.HEADER_PROPERTY_BLOB_SIZE
         offset += self.tl.OBJ_N_METHODS * self.tl.header.HEADER_FUNCTION_BLOB_SIZE
         n_signals = self.tl.OBJ_N_SIGNALS
@@ -362,30 +457,34 @@ class Class(GirNode, GirType):
         return result
 
     @cached_property
-    def parent(self):
+    def parent(self) -> T.Optional["Class"]:
         if entry := self.tl.OBJ_PARENT:
             return self.get_containing(Repository)._resolve_dir_entry(entry)
         else:
             return None
 
     @cached_property
-    def signature(self):
+    def signature(self) -> str:
+        assert self.container is not None
         result = f"class {self.container.name}.{self.name}"
         if self.parent is not None:
+            assert self.parent.container is not None
             result += f" : {self.parent.container.name}.{self.parent.name}"
         if len(self.implements):
-            result += " implements " + ", ".join([impl.full_name for impl in self.implements])
+            result += " implements " + ", ".join(
+                [impl.full_name for impl in self.implements]
+            )
         return result
 
     @cached_property
-    def properties(self):
-        return { p.name: p for p in self._enum_properties() }
+    def properties(self) -> T.Mapping[str, Property]:
+        return {p.name: p for p in self._enum_properties()}
 
     @cached_property
-    def signals(self):
-        return { s.name: s for s in self._enum_signals() }
+    def signals(self) -> T.Mapping[str, Signal]:
+        return {s.name: s for s in self._enum_signals()}
 
-    def assignable_to(self, other) -> bool:
+    def assignable_to(self, other: GirType) -> bool:
         if self == other:
             return True
         elif self.parent and self.parent.assignable_to(other):
@@ -397,7 +496,7 @@ class Class(GirNode, GirType):
 
             return False
 
-    def _enum_properties(self):
+    def _enum_properties(self) -> T.Iterable[Property]:
         yield from self.own_properties.values()
 
         if self.parent is not None:
@@ -406,7 +505,7 @@ class Class(GirNode, GirType):
         for impl in self.implements:
             yield from impl.properties.values()
 
-    def _enum_signals(self):
+    def _enum_signals(self) -> T.Iterable[Signal]:
         yield from self.own_signals.values()
 
         if self.parent is not None:
@@ -416,37 +515,91 @@ class Class(GirNode, GirType):
             yield from impl.signals.values()
 
 
-class EnumMember(GirNode):
-    def __init__(self, ns, tl: typelib.Typelib):
-        super().__init__(ns, tl)
+class TemplateType(GirType):
+    def __init__(self, name: str, parent: T.Optional[GirType]):
+        self._name = name
+        self.parent = parent
 
     @property
-    def value(self):
+    def name(self) -> str:
+        return self._name
+
+    @property
+    def full_name(self) -> str:
+        return self._name
+
+    @property
+    def glib_type_name(self) -> str:
+        return self._name
+
+    @cached_property
+    def properties(self) -> T.Mapping[str, Property]:
+        if not (isinstance(self.parent, Class) or isinstance(self.parent, Interface)):
+            return {}
+        else:
+            return self.parent.properties
+
+    @cached_property
+    def signals(self) -> T.Mapping[str, Signal]:
+        if not (isinstance(self.parent, Class) or isinstance(self.parent, Interface)):
+            return {}
+        else:
+            return self.parent.signals
+
+    def assignable_to(self, other: "GirType") -> bool:
+        if self == other:
+            return True
+        elif isinstance(other, Interface):
+            # we don't know the template type's interfaces, assume yes
+            return True
+        elif self.parent is None or isinstance(self.parent, ExternType):
+            return isinstance(other, Class)
+        else:
+            return self.parent.assignable_to(other)
+
+    @cached_property
+    def signature(self) -> str:
+        if self.parent is None:
+            return f"template {self.name}"
+        else:
+            return f"template {self.name} : {self.parent.full_name}"
+
+    @property
+    def incomplete(self) -> bool:
+        return True
+
+
+class EnumMember(GirNode):
+    def __init__(self, enum: "Enumeration", tl: typelib.Typelib) -> None:
+        super().__init__(enum, tl)
+
+    @property
+    def value(self) -> int:
         return self.tl.VALUE_VALUE
 
     @cached_property
-    def name(self):
+    def name(self) -> str:
         return self.tl.VALUE_NAME
 
     @cached_property
-    def nick(self):
+    def nick(self) -> str:
         return self.name.replace("_", "-")
 
     @property
-    def c_ident(self):
+    def c_ident(self) -> str:
         return self.tl.attr("c:identifier")
 
     @property
-    def signature(self):
+    def signature(self) -> str:
         return f"enum member {self.full_name} = {self.value}"
 
 
 class Enumeration(GirNode, GirType):
-    def __init__(self, ns, tl: typelib.Typelib):
+    def __init__(self, ns: "Namespace", tl: typelib.Typelib) -> None:
         super().__init__(ns, tl)
 
     @cached_property
-    def members(self):
+    def members(self) -> T.Dict[str, EnumMember]:
         members = {}
         n_values = self.tl.ENUM_N_VALUES
         values = self.tl.ENUM_VALUES
@@ -457,31 +610,43 @@ class Enumeration(GirNode, GirType):
         return members
 
     @property
-    def signature(self):
+    def signature(self) -> str:
         return f"enum {self.full_name}"
 
-    def assignable_to(self, type):
+    def assignable_to(self, type: GirType) -> bool:
+        return type == self
+
+
+class Boxed(GirNode, GirType):
+    def __init__(self, ns: "Namespace", tl: typelib.Typelib) -> None:
+        super().__init__(ns, tl)
+
+    @property
+    def signature(self) -> str:
+        return f"boxed {self.full_name}"
+
+    def assignable_to(self, type) -> bool:
         return type == self
 
 
 class Bitfield(Enumeration):
-    def __init__(self, ns, tl: typelib.Typelib):
+    def __init__(self, ns: "Namespace", tl: typelib.Typelib) -> None:
         super().__init__(ns, tl)
 
 
 class Namespace(GirNode):
-    def __init__(self, repo, tl: typelib.Typelib):
+    def __init__(self, repo: "Repository", tl: typelib.Typelib) -> None:
         super().__init__(repo, tl)
 
-        self.entries: T.Dict[str, GirNode] = {}
+        self.entries: T.Dict[str, GirType] = {}
 
-        n_local_entries = tl.HEADER_N_ENTRIES
-        directory = tl.HEADER_DIRECTORY
+        n_local_entries: int = tl.HEADER_N_ENTRIES
+        directory: typelib.Typelib = tl.HEADER_DIRECTORY
         for i in range(n_local_entries):
             entry = directory[i * tl.HEADER_ENTRY_BLOB_SIZE]
-            entry_name = entry.DIR_ENTRY_NAME
-            entry_type = entry.DIR_ENTRY_BLOB_TYPE
-            entry_blob = entry.DIR_ENTRY_OFFSET
+            entry_name: str = entry.DIR_ENTRY_NAME
+            entry_type: int = entry.DIR_ENTRY_BLOB_TYPE
+            entry_blob: typelib.Typelib = entry.DIR_ENTRY_OFFSET
 
             if entry_type == typelib.BLOB_TYPE_ENUM:
                 self.entries[entry_name] = Enumeration(self, entry_blob)
@@ -490,45 +655,59 @@ class Namespace(GirNode):
             elif entry_type == typelib.BLOB_TYPE_OBJECT:
                 self.entries[entry_name] = Class(self, entry_blob)
             elif entry_type == typelib.BLOB_TYPE_INTERFACE:
-                 self.entries[entry_name] = Interface(self, entry_blob)
+                self.entries[entry_name] = Interface(self, entry_blob)
+            elif (
+                entry_type == typelib.BLOB_TYPE_BOXED
+                or entry_type == typelib.BLOB_TYPE_STRUCT
+            ):
+                self.entries[entry_name] = Boxed(self, entry_blob)
 
     @cached_property
     def xml(self):
         return get_xml(self.name, self.version).get_elements("namespace")[0]
 
     @cached_property
-    def name(self):
+    def name(self) -> str:
         return self.tl.HEADER_NAMESPACE
 
     @cached_property
-    def version(self):
+    def version(self) -> str:
         return self.tl.HEADER_NSVERSION
 
     @property
-    def signature(self):
+    def signature(self) -> str:
         return f"namespace {self.name} {self.version}"
 
     @cached_property
-    def classes(self):
-        return { name: entry for name, entry in self.entries.items() if isinstance(entry, Class) }
+    def classes(self) -> T.Mapping[str, Class]:
+        return {
+            name: entry
+            for name, entry in self.entries.items()
+            if isinstance(entry, Class)
+        }
 
     @cached_property
-    def interfaces(self):
-        return { name: entry for name, entry in self.entries.items() if isinstance(entry, Interface) }
+    def interfaces(self) -> T.Mapping[str, Interface]:
+        return {
+            name: entry
+            for name, entry in self.entries.items()
+            if isinstance(entry, Interface)
+        }
 
-    def get_type(self, name):
-        """ Gets a type (class, interface, enum, etc.) from this namespace. """
+    def get_type(self, name) -> T.Optional[GirType]:
+        """Gets a type (class, interface, enum, etc.) from this namespace."""
         return self.entries.get(name)
 
-    def get_type_by_cname(self, cname: str):
-        """ Gets a type from this namespace by its C name. """
+    def get_type_by_cname(self, cname: str) -> T.Optional[GirType]:
+        """Gets a type from this namespace by its C name."""
         for item in self.entries.values():
             if hasattr(item, "cname") and item.cname == cname:
                 return item
+        return None
 
-    def lookup_type(self, type_name: str):
-        """ Looks up a type in the scope of this namespace (including in the
-        namespace's dependencies). """
+    def lookup_type(self, type_name: str) -> T.Optional[GirType]:
+        """Looks up a type in the scope of this namespace (including in the
+        namespace's dependencies)."""
 
         if type_name in _BASIC_TYPES:
             return _BASIC_TYPES[type_name]()
@@ -540,7 +719,7 @@ class Namespace(GirNode):
 
 
 class Repository(GirNode):
-    def __init__(self, tl: typelib.Typelib):
+    def __init__(self, tl: typelib.Typelib) -> None:
         super().__init__(None, tl)
 
         self.namespace = Namespace(self, tl)
@@ -548,25 +727,25 @@ class Repository(GirNode):
         if dependencies := tl[0x24].string:
             deps = [tuple(dep.split("-", 1)) for dep in dependencies.split("|")]
             try:
-                self.includes = { name: get_namespace(name, version) for name, version in deps }
+                self.includes = {
+                    name: get_namespace(name, version) for name, version in deps
+                }
             except:
                 raise CompilerBugError(f"Failed to load dependencies.")
         else:
             self.includes = {}
 
-    def get_type(self, name: str, ns: str) -> T.Optional[GirNode]:
+    def get_type(self, name: str, ns: str) -> T.Optional[GirType]:
         return self.lookup_namespace(ns).get_type(name)
 
-
-    def get_type_by_cname(self, name: str) -> T.Optional[GirNode]:
+    def get_type_by_cname(self, name: str) -> T.Optional[GirType]:
         for ns in [self.namespace, *self.includes.values()]:
             if type := ns.get_type_by_cname(name):
                 return type
         return None
 
-
     def lookup_namespace(self, ns: str):
-        """ Finds a namespace among this namespace's dependencies. """
+        """Finds a namespace among this namespace's dependencies."""
         if ns == self.namespace.name:
             return self.namespace
         else:
@@ -581,7 +760,7 @@ class Repository(GirNode):
             ns = dir_entry.DIR_ENTRY_NAMESPACE
             return self.lookup_namespace(ns).get_type(dir_entry.DIR_ENTRY_NAME)
 
-    def _resolve_type_id(self, type_id: int):
+    def _resolve_type_id(self, type_id: int) -> GirType:
         if type_id & 0xFFFFFF == 0:
             type_id = (type_id >> 27) & 0x1F
             # simple type
@@ -589,40 +768,62 @@ class Repository(GirNode):
                 return BoolType()
             elif type_id in [typelib.TYPE_FLOAT, typelib.TYPE_DOUBLE]:
                 return FloatType()
-            elif type_id in [typelib.TYPE_INT8, typelib.TYPE_INT16, typelib.TYPE_INT32, typelib.TYPE_INT64]:
+            elif type_id in [
+                typelib.TYPE_INT8,
+                typelib.TYPE_INT16,
+                typelib.TYPE_INT32,
+                typelib.TYPE_INT64,
+            ]:
                 return IntType()
-            elif type_id in [typelib.TYPE_UINT8, typelib.TYPE_UINT16, typelib.TYPE_UINT32, typelib.TYPE_UINT64]:
+            elif type_id in [
+                typelib.TYPE_UINT8,
+                typelib.TYPE_UINT16,
+                typelib.TYPE_UINT32,
+                typelib.TYPE_UINT64,
+            ]:
                 return UIntType()
             elif type_id == typelib.TYPE_UTF8:
                 return StringType()
+            elif type_id == typelib.TYPE_GTYPE:
+                return TypeType()
             else:
                 raise CompilerBugError("Unknown type ID", type_id)
         else:
-            return self._resolve_dir_entry(self.tl.header[type_id].INTERFACE_TYPE_INTERFACE)
-
+            blob = self.tl.header[type_id]
+            if blob.TYPE_BLOB_TAG == typelib.TYPE_INTERFACE:
+                return self._resolve_dir_entry(
+                    self.tl.header[type_id].TYPE_BLOB_INTERFACE
+                )
+            elif blob.TYPE_BLOB_TAG == typelib.TYPE_ARRAY:
+                return ArrayType(self._resolve_type_id(blob.TYPE_BLOB_ARRAY_INNER))
+            else:
+                raise CompilerBugError(f"{blob.TYPE_BLOB_TAG}")
 
 
 class GirContext:
     def __init__(self):
         self.namespaces = {}
-
+        self.not_found_namespaces: T.Set[str] = set()
 
     def add_namespace(self, namespace: Namespace):
         other = self.namespaces.get(namespace.name)
         if other is not None and other.version != namespace.version:
-            raise CompileError(f"Namespace {namespace.name}-{namespace.version} can't be imported because version {other.version} was imported earlier")
+            raise CompileError(
+                f"Namespace {namespace.name}-{namespace.version} can't be imported because version {other.version} was imported earlier"
+            )
 
         self.namespaces[namespace.name] = namespace
 
-
-    def get_type_by_cname(self, name: str) -> T.Optional[GirNode]:
+    def get_type_by_cname(self, name: str) -> T.Optional[GirType]:
         for ns in self.namespaces.values():
             if type := ns.get_type_by_cname(name):
                 return type
         return None
 
+    def get_type(self, name: str, ns: str) -> T.Optional[GirType]:
+        if ns is None and name in _BASIC_TYPES:
+            return _BASIC_TYPES[name]()
 
-    def get_type(self, name: str, ns: str) -> T.Optional[GirNode]:
         ns = ns or "Gtk"
 
         if ns not in self.namespaces:
@@ -630,7 +831,6 @@ class GirContext:
 
         return self.namespaces[ns].get_type(name)
 
-
     def get_class(self, name: str, ns: str) -> T.Optional[Class]:
         type = self.get_type(name, ns)
         if isinstance(type, Class):
@@ -638,21 +838,20 @@ class GirContext:
         else:
             return None
 
-
-    def validate_ns(self, ns: str):
-        """ Raises an exception if there is a problem looking up the given
-        namespace. """
+    def validate_ns(self, ns: str) -> None:
+        """Raises an exception if there is a problem looking up the given
+        namespace."""
 
         ns = ns or "Gtk"
 
-        if ns not in self.namespaces:
+        if ns not in self.namespaces and ns not in self.not_found_namespaces:
             raise CompileError(
                 f"Namespace {ns} was not imported",
                 did_you_mean=(ns, self.namespaces.keys()),
             )
 
-    def validate_type(self, name: str, ns: str):
-        """ Raises an exception if there is a problem looking up the given type. """
+    def validate_type(self, name: str, ns: str) -> None:
+        """Raises an exception if there is a problem looking up the given type."""
 
         self.validate_ns(ns)
 
