@@ -25,8 +25,18 @@ from .common import *
 class StyleClass(AstNode):
     grammar = UseQuoted("name")
 
+    @property
+    def name(self) -> str:
+        return self.tokens["name"]
 
-class Styles(AstNode):
+    @validate("name")
+    def unique_in_parent(self):
+        self.validate_unique_in_parent(
+            f"Duplicate style class '{self.name}'", lambda x: x.name == self.name
+        )
+
+
+class ExtStyles(AstNode):
     grammar = [
         Keyword("styles"),
         "[",
@@ -49,13 +59,14 @@ class Styles(AstNode):
     matches=new_statement_patterns,
 )
 def style_completer(ast_node, match_variables):
-    yield Completion("styles", CompletionItemKind.Keyword, snippet="styles [\"$0\"]")
+    yield Completion("styles", CompletionItemKind.Keyword, snippet='styles ["$0"]')
 
 
 @decompiler("style")
 def decompile_style(ctx, gir):
     ctx.print(f"styles [")
 
+
 @decompiler("class")
 def decompile_style_class(ctx, gir, name):
     ctx.print(f'"{name}",')
